using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Windows.Media;
using Styx;
using Styx.Common;
using Styx.Plugins;
using Styx.WoWInternals;

namespace NuokRealmHop
{
    internal class Plugin : HBPlugin
    {
        public static List<CategoryInfo> AvailableCategories = new List<CategoryInfo>();
        private static CategoryInfo _lastSearchCategory;
        private static bool _isWaitingForResults;

        private static readonly List<SearchResultInfo> SearchResults = new List<SearchResultInfo>();

        private static readonly Dictionary<CategoryInfo, DateTime> LastUpdate =
            new Dictionary<CategoryInfo, DateTime>();

        private static readonly Dictionary<string, DateTime> ServerTracker = new Dictionary<string, DateTime>();
        private static List<ApplicationInfo> _applicationInfos = new List<ApplicationInfo>();
        private static readonly Stopwatch Sw = new Stopwatch();
        private static readonly Stopwatch JumpTimer = new Stopwatch();

        private static bool _resultsUpdated;

        private static int _pulseCounter;

        private static readonly HashSet<int> HasAppliedId = new HashSet<int>();

        public override string Name => "NuokRealmHop";
        public override string Author => "Nuok";
        public override Version Version => new Version(3,0);
        public override bool WantButton => true;

        private static void Log(string format, params object[] args)
        {
            Logging.WriteDiagnostic(Colors.Azure, "NuokRealmHop: " + format, args);
        }

        public override void OnEnable()
        {
            Log($"{Name} {Version} Enabled");
            LfgList.LFGLiftFrame_Show();
            Lua.Events.AttachEvent("LFG_LIST_SEARCH_FAILED", SearchFailed);
            Lua.Events.AttachEvent("LFG_LIST_SEARCH_RESULTS_RECEIVED", ResultsRecived);
            Lua.Events.AttachEvent("LFG_LIST_SEARCH_RESULT_UPDATED", ResultsUpdated);
            AvailableCategories = LfgList.GetAvailableCategories;

            _isWaitingForResults = false;
            _resultsUpdated = false;
            SearchResults.Clear();
            LastUpdate.Clear();
            _applicationInfos.Clear();

            var aCat = AvailableCategories.Aggregate("", (current, c) => current + $"{c.Name} ");
            Log(aCat);
        }

        private void SearchFailed(object sender, LuaEventArgs args)
        {
            Log($"LFG_LIST_SEARCH_FAILED {_lastSearchCategory}");
        }


        private void ResultsUpdated(object sender, LuaEventArgs args)
        {
            if (NuokRealmHopSettings.Instance.DebugDetailed)
                Log("LFG_LIST_SEARCH_RESULT_UPDATED");
            _resultsUpdated = true;
            _isWaitingForResults = false;
        }

        public override void OnDisable()
        {
            Log($"{Name} {Version} Disabled");
            Lua.Events.DetachEvent("LFG_LIST_SEARCH_FAILED", SearchFailed);
            Lua.Events.DetachEvent("LFG_LIST_SEARCH_RESULTS_RECEIVED", ResultsRecived);
            Lua.Events.DetachEvent("LFG_LIST_SEARCH_RESULT_UPDATED", ResultsUpdated);
            SearchResults.Clear();
            _pulseCounter = 0;

            _isWaitingForResults = false;
            _resultsUpdated = false;
            SearchResults.Clear();
            LastUpdate.Clear();
            _applicationInfos.Clear();
            Sw.Stop();
            JumpTimer.Stop();
        }

        public override void OnButtonPress()
        {
            string logTag = $"{Name} v{Version}";
            var window = new ConfigWindow(Name, logTag, Author, 350, 400);
            window.Show();
            window.Closed += ConfigWindowOnClosed;
        }

        private void ConfigWindowOnClosed(object sender, EventArgs e)
        {
            NuokRealmHopSettings.Instance.Save();
            Log("Settings Saved");
        }

        private void ResultsRecived(object sender, LuaEventArgs args)
        {
            using (new PerformanceLogger(NuokRealmHopSettings.Instance.DebugPerformace, "ResultsRecived"))
            {
                Log("LFG_LIST_SEARCH_RESULTS_RECEIVED");
                var results = LfgList.GetSearchResults;
                SearchResults.AddRange(results);
                if (LastUpdate.ContainsKey(_lastSearchCategory))
                    LastUpdate[_lastSearchCategory] = DateTime.Now;
                else
                    LastUpdate.Add(_lastSearchCategory, DateTime.Now);
            }
        }

        private static bool CategoryEnabled(CategoryInfo category)
        {
            var s = NuokRealmHopSettings.Instance;
            switch (category.Category)
            {
                case LFGCategory.Arenas:
                    return s.Arenas;
                case LFGCategory.Ashran:
                    return s.Ashran;
                case LFGCategory.Battlegrounds:
                    return s.Battlegrounds;
                case LFGCategory.Custom:
                    return s.Custom;
                case LFGCategory.Dungeons:
                    return s.Dungeons;
                case LFGCategory.Questing:
                    return s.Questing;
                case LFGCategory.Raids:
                    return s.Raids;
                case LFGCategory.Scenarios:
                    return s.Scenarios;
                case LFGCategory.Skirmishes:
                    return s.Skirmishes;
            }
            return false;
        }

        public override void Pulse()
        {
            var s = NuokRealmHopSettings.Instance;
            _pulseCounter++;
            if (_pulseCounter < s.DelayPulse) return;
            _pulseCounter = 0;

            using (new PerformanceLogger(s.DebugPerformace, "Pulse"))
            {
                if (AvailableCategories.Count < 1) AvailableCategories = LfgList.GetAvailableCategories;
                var cRealm = LfgList.GetRealmName;

                if (cRealm == null)
                {
                    if (s.DebugDetailed)
                        Log("GetRealmName returned null");
                    return;
                }
                if (ServerTracker.ContainsKey(cRealm))
                    ServerTracker[cRealm] = DateTime.Now;
                else
                    ServerTracker.Add(cRealm, DateTime.Now);

                //if were in a party lets check if we should leave after a certain time or continue on
                if (StyxWoW.Me.GroupInfo.IsInParty || StyxWoW.Me.GroupInfo.IsInRaid)
                {
                    if (StyxWoW.Me.GroupInfo.GroupLeader == StyxWoW.Me && s.LeaderLeaver)
                    {
                        Log("Leaving the group as your its leader");
                        LfgList.LeaveGroup();
                        return;
                    }
                    if (s.LeaveAfterMin <= 0) return;
                    if (!JumpTimer.IsRunning)
                    {
                        JumpTimer.Restart();
                        return;
                    }
                    if (!(JumpTimer.Elapsed.TotalMinutes > s.LeaveAfterMin)) return;
                    Log("Leaving group to Jump Realm");
                    if (!LfgList.LeaveGroup()) return;
                    JumpTimer.Stop();
                    _applicationInfos = new List<ApplicationInfo>();
                    SearchResults.Clear();
                    return;
                }

                if (_resultsUpdated)
                {
                    _applicationInfos = LfgList.GetApplicationInfos;
                    Log($"LFG_LIST_SEARCH_RESULT_UPDATED found {_applicationInfos.Count} applications");
                    if (_applicationInfos.Any(a => a.Status == ApplicationState.Invited))
                    {
                        LfgList.AcceptPopUps();
                    }
                    _resultsUpdated = false;
                }

                var applications = LfgList.NumActiveAppliications;

                if (s.DebugDetailed)
                    Log($"NumActiveAppliications: {applications}");

                if (applications >= 5) return;

                foreach (var r in from r in SearchResults
                    where !HasAppliedId.Contains(r.Id)
                    where !r.IsDelisted
                    where !_applicationInfos.Exists(a => a.Id == r.Id)
                    where !(r.iLevel > StyxWoW.Me.AverageItemLevelEquipped)
                    where !s.AutoInvite || r.AutoInvite
                    where r.PlayerCount <= s.MaxPartySize
                    where
                        !ServerTracker.ContainsKey(r.Realm) ||
                        !(DateTime.Now.Subtract(ServerTracker[r.Realm]).TotalMinutes < s.RealmBlacklist)
                    select r)
                {
                    HasAppliedId.Add(r.Id);
                    r.ApplyToGroup();
                    applications = LfgList.NumActiveAppliications;
                    if (applications >= 5) break;
                }


                //If were searching wait until it finishes or times out
                if (_isWaitingForResults && _lastSearchCategory != null)
                {
                    if (!Sw.IsRunning) Sw.Restart();
                    if (!(Sw.Elapsed.TotalSeconds > 5)) return;
                    Log($"{_lastSearchCategory} search has timed out, researching");
                    if (LastUpdate.ContainsKey(_lastSearchCategory))
                        LastUpdate.Remove(_lastSearchCategory);
                    _lastSearchCategory = null;
                    Sw.Stop();
                    return;
                }
                Sw.Stop();

                //Start a new search if not searching
                foreach (var cat in AvailableCategories.Where(cat => _lastSearchCategory != cat)
                    .Where(
                        cat => CategoryEnabled(cat) &&
                               (!LastUpdate.ContainsKey(cat) ||
                                !(DateTime.Now.Subtract(LastUpdate[cat]).TotalSeconds < 30))))
                {
                    cat.SearchCatergory();
                    _lastSearchCategory = cat;
                    _isWaitingForResults = true;
                    break;
                }
            }
        }
    }
}